/*
 * Copyright 2016 Samsung Electronics S.LSI Co. LTD
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @file        Exynos_OSAL_Tizen.c
 * @brief
 * @author      Taehwan Kim (t_h.kim@samsung.com)
 * @version     1.0.0
 * @history
 *   2016.06.13 : Create
 */

#include <stdio.h>
#include <stdlib.h>

#include "Exynos_OSAL_Semaphore.h"
#include "Exynos_OMX_Baseport.h"
#include "Exynos_OMX_Basecomponent.h"
#include "Exynos_OMX_Macros.h"
#include "Exynos_OSAL_Platform.h"
#include "Exynos_OSAL_ETC.h"
#include "exynos_format.h"

#include "ExynosVideoApi.h"

#ifdef TIZEN_FEATURE_TIZEN_ALLOCATOR
#include <tbm_bufmgr.h>
#include <tbm_surface.h>
#include <tbm_surface_internal.h>
#else
#include <mmf/mm_types.h>
#endif

#undef  EXYNOS_LOG_TAG
#define EXYNOS_LOG_TAG    "Exynos_OSAL_Tizen"
//#define EXYNOS_LOG_OFF
#include "Exynos_OSAL_Log.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct __EnableGemBuffersParams
{
  OMX_U32           nSize;
  OMX_VERSIONTYPE   nVersion;
  OMX_U32           nPortIndex;
  OMX_BOOL          bStoreMetaData;
} EnableGemBuffersParams;

static OMX_ERRORTYPE setBufferProcessTypeForDecoder(EXYNOS_OMX_BASEPORT *pExynosPort)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    FunctionIn();

    if (pExynosPort == NULL) {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] invalid parameter", __FUNCTION__);
        ret = OMX_ErrorUndefined;
        goto EXIT;
    }

    if ((pExynosPort->bufferProcessType == BUFFER_COPY) &&
        (pExynosPort->eMetaDataType != METADATA_TYPE_DISABLED)) {
        int i;
        if (pExynosPort->supportFormat == NULL) {
            ret = OMX_ErrorUndefined;
            Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] supported format is empty", __FUNCTION__);
            goto EXIT;
        }

        pExynosPort->portDefinition.format.video.eColorFormat = OMX_COLOR_FormatYUV420SemiPlanar;
        /* check to support NV12T format */
        for (i = 0; pExynosPort->supportFormat[i] != OMX_COLOR_FormatUnused; i++) {
            /* prefer to use NV12T */
            if (pExynosPort->supportFormat[i] == (OMX_COLOR_FORMATTYPE)OMX_SEC_COLOR_FormatNV12Tiled) {
                pExynosPort->portDefinition.format.video.eColorFormat = (OMX_COLOR_FORMATTYPE)OMX_SEC_COLOR_FormatNV12Tiled;
                break;
            }
        }
        pExynosPort->bufferProcessType = BUFFER_SHARE;

        Exynos_OSAL_Log(EXYNOS_LOG_ESSENTIAL, "[%s] output buffer sharing mode is on (%x)",
                                                __FUNCTION__, pExynosPort->portDefinition.format.video.eColorFormat);
    }

EXIT:
    FunctionOut();

    return ret;
}

static void resetBufferProcessTypeForDecoder(EXYNOS_OMX_BASEPORT *pExynosPort)
{
    FunctionIn();

    if (pExynosPort == NULL) {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] invalid parameter", __FUNCTION__);
        goto EXIT;
    }

    pExynosPort->bufferProcessType = BUFFER_COPY;
    pExynosPort->portDefinition.format.video.eColorFormat = OMX_COLOR_FormatYUV420Planar;

EXIT:
    FunctionOut();

    return;
}

OMX_ERRORTYPE Exynos_OSAL_GetParameter(
    OMX_IN OMX_HANDLETYPE hComponent,
    OMX_IN OMX_INDEXTYPE  nIndex,
    OMX_INOUT OMX_PTR     pComponentParameterStructure)
{
    OMX_ERRORTYPE                ret                = OMX_ErrorNone;
    OMX_COMPONENTTYPE           *pOMXComponent      = NULL;
    EXYNOS_OMX_BASECOMPONENT    *pExynosComponent   = NULL;

    FunctionIn();

    if (hComponent == NULL) {
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    pOMXComponent = (OMX_COMPONENTTYPE *)hComponent;
    ret = Exynos_OMX_Check_SizeVersion(pOMXComponent, sizeof(OMX_COMPONENTTYPE));
    if (ret != OMX_ErrorNone)
        goto EXIT;

    if (pOMXComponent->pComponentPrivate == NULL) {
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    pExynosComponent = (EXYNOS_OMX_BASECOMPONENT *)pOMXComponent->pComponentPrivate;
    if (pExynosComponent->currentState == OMX_StateInvalid) {
        ret = OMX_ErrorInvalidState;
        goto EXIT;
    }

    if (pComponentParameterStructure == NULL) {
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    switch ((int)nIndex) {
    case OMX_IndexParamPortDefinition:
    {
        OMX_PARAM_PORTDEFINITIONTYPE    *pPortDef       = (OMX_PARAM_PORTDEFINITIONTYPE *)pComponentParameterStructure;
        EXYNOS_OMX_BASEPORT             *pExynosPort    = &pExynosComponent->pExynosPort[pPortDef->nPortIndex];

        if (pExynosPort->eMetaDataType == METADATA_TYPE_DATA)
#ifdef TIZEN_FEATURE_TIZEN_ALLOCATOR
            pPortDef->nBufferSize = (ALIGN(pExynosPort->portDefinition.format.video.nFrameWidth, 16) *
                                     ALIGN(pExynosPort->portDefinition.format.video.nFrameHeight, 16) * 3) / 2;
#else
            pPortDef->nBufferSize = sizeof(MMVideoBuffer);
#endif
    }
        break;
    default:
    {
        ret = OMX_ErrorUnsupportedIndex;
        goto EXIT;
    }
        break;
    }

EXIT:
    FunctionOut();

    return ret;
}

OMX_ERRORTYPE Exynos_OSAL_SetParameter(
    OMX_IN OMX_HANDLETYPE hComponent,
    OMX_IN OMX_INDEXTYPE  nIndex,
    OMX_IN OMX_PTR        pComponentParameterStructure)
{
    OMX_ERRORTYPE                ret                = OMX_ErrorNone;
    OMX_COMPONENTTYPE           *pOMXComponent      = NULL;
    EXYNOS_OMX_BASECOMPONENT    *pExynosComponent   = NULL;

    FunctionIn();

    if (hComponent == NULL) {
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    pOMXComponent = (OMX_COMPONENTTYPE *)hComponent;
    ret = Exynos_OMX_Check_SizeVersion(pOMXComponent, sizeof(OMX_COMPONENTTYPE));
    if (ret != OMX_ErrorNone)
        goto EXIT;

    if (pOMXComponent->pComponentPrivate == NULL) {
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    pExynosComponent = (EXYNOS_OMX_BASECOMPONENT *)pOMXComponent->pComponentPrivate;
    if (pExynosComponent->currentState == OMX_StateInvalid) {
        ret = OMX_ErrorInvalidState;
        goto EXIT;
    }

    if (pComponentParameterStructure == NULL) {
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    Exynos_OSAL_Log(EXYNOS_LOG_ESSENTIAL, "[%p][%s] index = 0x%x", pExynosComponent, __FUNCTION__, nIndex);
    switch ((int)nIndex) {
    case OMX_IndexParamStoreMetaDataBuffer:
    {
        EnableGemBuffersParams *pGBParams   = (EnableGemBuffersParams *)pComponentParameterStructure;
        OMX_U32                 nPortIndex  = pGBParams->nPortIndex;
        EXYNOS_OMX_BASEPORT    *pExynosPort = NULL;

        ret = Exynos_OMX_Check_SizeVersion(pGBParams, sizeof(EnableGemBuffersParams));
        if (ret != OMX_ErrorNone) {
            Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%p][%s] Failed to Check_SizeVersion", pExynosComponent, __FUNCTION__);
            goto EXIT;
        }

        if (nPortIndex >= pExynosComponent->portParam.nPorts) {
            ret = OMX_ErrorBadPortIndex;
            goto EXIT;
        }

        if ((pExynosComponent->codecType == HW_VIDEO_DEC_SECURE_CODEC) ||
            (pExynosComponent->codecType == HW_VIDEO_DEC_CODEC)) {
            /* decoder */
            if (nPortIndex != OUTPUT_PORT_INDEX) {
                ret = OMX_ErrorNotImplemented;
                goto EXIT;
            }
        } else {
            /* encoder */
            if (nPortIndex != INPUT_PORT_INDEX) {
                ret = OMX_ErrorNotImplemented;
                goto EXIT;
            }
        }

        pExynosPort = &pExynosComponent->pExynosPort[nPortIndex];
        if (CHECK_PORT_TUNNELED(pExynosPort) && CHECK_PORT_BUFFER_SUPPLIER(pExynosPort)) {
            ret = OMX_ErrorBadPortIndex;
            goto EXIT;
        }

        Exynos_OSAL_Log(EXYNOS_LOG_ESSENTIAL, "[%p][%s][META] %s port: metadata mode(%x) -> %s",
                                        pExynosComponent, __FUNCTION__,
                                        (nPortIndex == INPUT_PORT_INDEX)? "input":"output",
                                        pExynosPort->eMetaDataType,
                                        (pGBParams->bStoreMetaData == OMX_TRUE)? "enable":"disable");

        if ((pGBParams->bStoreMetaData == OMX_TRUE) &&
            (pExynosPort->eMetaDataType == METADATA_TYPE_DISABLED)) {
            pExynosPort->eMetaDataType = METADATA_TYPE_DATA;
        }

        /* in case of decoder output */
        if ((nPortIndex == OUTPUT_PORT_INDEX) &&
            ((pExynosComponent->codecType == HW_VIDEO_DEC_CODEC) ||
                (pExynosComponent->codecType == HW_VIDEO_DEC_SECURE_CODEC))) {
            if (pGBParams->bStoreMetaData == OMX_TRUE) {
                ret = setBufferProcessTypeForDecoder(pExynosPort);
                if (ret != OMX_ErrorNone)
                    goto EXIT;
            } else if ((pGBParams->bStoreMetaData == OMX_FALSE) &&
                       (pExynosPort->eMetaDataType != METADATA_TYPE_DISABLED)) {
                resetBufferProcessTypeForDecoder(pExynosPort);
            }
        }

        if ((pGBParams->bStoreMetaData == OMX_FALSE) &&
            (pExynosPort->eMetaDataType != METADATA_TYPE_DISABLED)) {
            pExynosPort->eMetaDataType = METADATA_TYPE_DISABLED;
        }
    }
        break;
    default:
    {
        ret = OMX_ErrorUnsupportedIndex;
        goto EXIT;
    }
        break;
    }

EXIT:
    FunctionOut();

    return ret;
}

/*
 * meta data contains the following data format.
 * 1) METADATA_TYPE_DATA
 * --------------------------------------------------------------------
 * | MMVideoBuffer                                                    | [DEC] out, [ENC] in
 * --------------------------------------------------------------------
 */
OMX_ERRORTYPE Exynos_OSAL_LockMetaData(
    OMX_IN OMX_PTR                          pBuffer,
    OMX_IN EXYNOS_OMX_LOCK_RANGE            range,
    OMX_OUT OMX_U32                        *pStride,
    OMX_OUT EXYNOS_OMX_MULTIPLANE_BUFFER   *pBufferInfo,
    OMX_IN EXYNOS_METADATA_TYPE             eMetaType)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    FunctionIn();

    if (pBuffer == NULL) {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] invalid parameter", __FUNCTION__);
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    /* not used yet */

EXIT:
    FunctionOut();

    return ret;
}

OMX_ERRORTYPE Exynos_OSAL_UnlockMetaData(
    OMX_IN OMX_PTR              pBuffer,
    OMX_IN EXYNOS_METADATA_TYPE eMetaType)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    FunctionIn();

    if (pBuffer == NULL) {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] invalid parameter", __FUNCTION__);
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    /* not used yet */

EXIT:
    FunctionOut();

    return ret;
}

OMX_ERRORTYPE Exynos_OSAL_GetInfoFromMetaData(
    OMX_IN OMX_PTR                          pBuffer,
    OMX_OUT EXYNOS_OMX_MULTIPLANE_BUFFER   *pBufferInfo,
    OMX_IN EXYNOS_METADATA_TYPE             eMetaDataType)
{
    OMX_ERRORTYPE    ret            = OMX_ErrorNone;
#ifdef TIZEN_FEATURE_TIZEN_ALLOCATOR
    tbm_surface_h surface;
    tbm_bo bo[2];
#else
    MMVideoBuffer   *pMetaBuffer    = NULL;
#endif

    int i;

    FunctionIn();

    if ((pBuffer == NULL) ||
        (pBufferInfo == NULL) ||
        (eMetaDataType == METADATA_TYPE_DISABLED)) {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] invalid parameter", __FUNCTION__);
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    if (eMetaDataType == METADATA_TYPE_DATA) {
#ifdef TIZEN_FEATURE_TIZEN_ALLOCATOR
        surface = (tbm_surface_h)pBuffer;

        bo[0] = tbm_surface_internal_get_bo(surface, 0);
        bo[1] = tbm_surface_internal_get_bo(surface, 1);

        for (i = 0; i < 2; i++) {
            pBufferInfo->fd[i]   = (unsigned long)(tbm_bo_get_handle (bo[i], TBM_DEVICE_MM).u32);
            pBufferInfo->addr[i] = (OMX_PTR)(tbm_bo_get_handle (bo[i], TBM_DEVICE_CPU).ptr);
#else
        pMetaBuffer = (MMVideoBuffer *)pBuffer;

        for (i = 0; i < pMetaBuffer->plane_num; i++) {
            pBufferInfo->fd[i]   = (unsigned long)(pMetaBuffer->handle.dmabuf_fd[i]);
            pBufferInfo->addr[i] = (OMX_PTR)(pMetaBuffer->data[i]);
#endif

            Exynos_OSAL_Log(EXYNOS_LOG_ESSENTIAL, "[%s] Plane[%d]: FD(%u), VA(%p)",
                                                    __FUNCTION__, i, pBufferInfo->fd[i], pBufferInfo->addr[i]);
        }
    } else {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] not implemented yet(meta type:0x%x)", __FUNCTION__, eMetaDataType);
        ret = OMX_ErrorNotImplemented;
        goto EXIT;
    }

EXIT:
    FunctionOut();

    return ret;
}

// Porting exynos 9110, code from SM-R765(7270), But not used
OMX_ERRORTYPE Exynos_OSAL_SetDataLengthToMetaData(
    OMX_IN OMX_BYTE             pBuffer,
    OMX_IN OMX_U32              dataLength,
    OMX_IN EXYNOS_METADATA_TYPE eMetaDataType)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    FunctionIn();

    if (pBuffer == NULL) {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] invalid parameter", __FUNCTION__);
        ret = OMX_ErrorBadParameter;
        goto EXIT;
    }

    /* not used yet */
    /* This is for that MetaData is configured by component */

EXIT:
    FunctionOut();

    return ret;
}

OMX_PTR Exynos_OSAL_AllocMetaDataBuffer(
    OMX_HANDLETYPE          hSharedMemory,
    EXYNOS_METADATA_TYPE    eMetaDataType,
    OMX_U32                 nSizeBytes,
    MEMORY_TYPE             eMemoryType)
{
    OMX_PTR pBuffer = NULL;

    FunctionIn();

    if (eMetaDataType == METADATA_TYPE_DATA) {
#ifdef TIZEN_FEATURE_TIZEN_ALLOCATOR
        pBuffer = Exynos_OSAL_Malloc(sizeof(tbm_surface_h));
#else
        pBuffer = Exynos_OSAL_Malloc(sizeof(MMVideoBuffer));
#endif
        if (pBuffer == NULL) {
            Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] Failed to allocate metadata buffer", __FUNCTION__);
            goto EXIT;
        }
    } else {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] not implemented yet(meta type:0x%x)", __FUNCTION__, eMetaDataType);
    }

EXIT:
    FunctionOut();

    return pBuffer;
}

OMX_ERRORTYPE Exynos_OSAL_FreeMetaDataBuffer(
    OMX_HANDLETYPE          hSharedMemory,
    EXYNOS_METADATA_TYPE    eMetaDataType,
    OMX_PTR                 pTempBuffer)
{
    OMX_ERRORTYPE ret = OMX_ErrorNone;

    FunctionIn();

    if (eMetaDataType == METADATA_TYPE_DATA) {
        Exynos_OSAL_Free(pTempBuffer);
        ret = OMX_ErrorNone;
    } else {
        Exynos_OSAL_Log(EXYNOS_LOG_ERROR, "[%s] not implemented yet(meta type:0x%x)", __FUNCTION__, eMetaDataType);
        ret = OMX_ErrorNotImplemented;
        goto EXIT;
    }
EXIT:
    FunctionOut();

    return ret;
}

#ifdef __cplusplus
}
#endif
